import numpy as np

import pybamm


def graphite_diffusivity_PeymanMPM(sto, T):
    """
    Graphite diffusivity as a function of stoichiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from Peyman MPM.

    References
    ----------
    .. [1] John Newman, Dualfoil

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stoichiometry
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_ref = 5.0 * 10 ** (-15)
    E_D_s = 42770
    arrhenius = np.exp(E_D_s / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius


def graphite_ocp_PeymanMPM(sto):
    """
    Graphite Open-circuit Potential (OCP) as a function of the
    stoichiometry. The fit is taken from Peyman MPM [1].

    References
    ----------
    .. [1] Peyman Mohtat et al, MPM (to be submitted)
    """

    u_eq = (
        0.063
        + 0.8 * np.exp(-75 * (sto + 0.001))
        - 0.0120 * np.tanh((sto - 0.127) / 0.016)
        - 0.0118 * np.tanh((sto - 0.155) / 0.016)
        - 0.0035 * np.tanh((sto - 0.220) / 0.020)
        - 0.0095 * np.tanh((sto - 0.190) / 0.013)
        - 0.0145 * np.tanh((sto - 0.490) / 0.020)
        - 0.0800 * np.tanh((sto - 1.030) / 0.055)
    )

    return u_eq


def graphite_electrolyte_exchange_current_density_PeymanMPM(c_e, c_s_surf, c_s_max, T):
    """
    Exchange-current density for Butler-Volmer reactions between graphite and LiPF6 in
    EC:DMC.
    Check the unit of Reaction rate constant k0 is from Peyman MPM.

    References
    ----------
    .. [2] John Newman, Dualfoil

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 1.061 * 10 ** (-6)  # unit has been converted
    # units are (A/m2)(m3/mol)**1.5 - includes ref concentrations
    E_r = 37480
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def graphite_entropic_change_PeymanMPM(sto):
    """
    Graphite entropic change in open-circuit potential (OCP) at a temperature of
    298.15K as a function of the stoichiometry taken from [1]

    References
    ----------
    .. [1] K.E. Thomas, J. Newman, "Heats of mixing and entropy in porous insertion
           electrode", J. of Power Sources 119 (2003) 844-849

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
        stoichiometry of material (li-fraction)

    """

    du_dT = 10 ** (-3) * (
        0.28
        - 1.56 * sto
        - 8.92 * sto ** (2)
        + 57.21 * sto ** (3)
        - 110.7 * sto ** (4)
        + 90.71 * sto ** (5)
        - 27.14 * sto ** (6)
    )

    return du_dT


def NMC_diffusivity_PeymanMPM(sto, T):
    """
    NMC diffusivity as a function of stoichiometry, in this case the
    diffusivity is taken to be a constant. The value is taken from Peyman MPM.

    References
    ----------
    .. [1] John Newman, Dualfoil

    Parameters
    ----------
    sto: :class:`pybamm.Symbol`
        Electrode stoichiometry
    T: :class:`pybamm.Symbol`
        Dimensional temperature

    Returns
    -------
    :class:`pybamm.Symbol`
        Solid diffusivity
    """

    D_ref = 8 * 10 ** (-15)
    E_D_s = 18550
    arrhenius = np.exp(E_D_s / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_ref * arrhenius


def NMC_ocp_PeymanMPM(sto):
    """
    Nickel Managanese Cobalt Oxide (NMC) Open-circuit Potential (OCP) as a
    function of the stoichiometry. The fit is taken from Peyman MPM.

    References
    ----------
    Peyman MPM manuscript (to be submitted)

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
       stoichiometry of material (li-fraction)

    """

    u_eq = (
        4.3452
        - 1.6518 * sto
        + 1.6225 * (sto**2)
        - 2.0843 * (sto**3)
        + 3.5146 * (sto**4)
        - 2.2166 * (sto**5)
        - 0.5623e-4 * np.exp(109.451 * sto - 100.006)
    )

    return u_eq


def NMC_electrolyte_exchange_current_density_PeymanMPM(c_e, c_s_surf, c_s_max, T):
    """
    Exchange-current density for Butler-Volmer reactions between NMC and LiPF6 in
    EC:DMC.

    References
    ----------
    .. Peyman MPM manuscript (to be submitted)

    Parameters
    ----------
    c_e : :class:`pybamm.Symbol`
        Electrolyte concentration [mol.m-3]
    c_s_surf : :class:`pybamm.Symbol`
        Particle concentration [mol.m-3]
    c_s_max : :class:`pybamm.Symbol`
        Maximum particle concentration [mol.m-3]
    T : :class:`pybamm.Symbol`
        Temperature [K]

    Returns
    -------
    :class:`pybamm.Symbol`
        Exchange-current density [A.m-2]
    """
    m_ref = 4.824 * 10 ** (-6)  # (A/m2)(m3/mol)**1.5 - includes ref concentrations
    E_r = 39570
    arrhenius = np.exp(E_r / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return m_ref * arrhenius * c_e**0.5 * c_s_surf**0.5 * (c_s_max - c_s_surf) ** 0.5


def NMC_entropic_change_PeymanMPM(sto):
    """
    Nickel Manganese Cobalt (NMC) entropic change in open-circuit potential (OCP) at
    a temperature of 298.15K as a function of the OCP. The fit is taken from [1].

    References
    ----------
    .. [1] W. Le, I. Belharouak, D. Vissers, K. Amine, "In situ thermal study of
    li1+ x [ni1/ 3co1/ 3mn1/ 3] 1- x o2 using isothermal micro-clorimetric
    techniques",
    J. of the Electrochemical Society 153 (11) (2006) A2147–A2151.

    Parameters
    ----------
    sto : :class:`pybamm.Symbol`
        stoichiometry of material (li-fraction)

    """

    # Since the equation uses the OCP at each stoichiometry as input,
    # we need OCP function here

    u_eq = (
        4.3452
        - 1.6518 * sto
        + 1.6225 * sto**2
        - 2.0843 * sto**3
        + 3.5146 * sto**4
        - 0.5623 * 10 ** (-4) * np.exp(109.451 * sto - 100.006)
    )

    du_dT = (-800 + 779 * u_eq - 284 * u_eq**2 + 46 * u_eq**3 - 2.8 * u_eq**4) * 10 ** (
        -3
    )

    return du_dT


def electrolyte_diffusivity_PeymanMPM(c_e, T):
    """
    Diffusivity of LiPF6 in EC:DMC as a function of ion concentration. The original data
    is from [1]. The fit from Dualfoil [2].

    References
    ----------
    .. [1] C Capiglia et al. 7Li and 19F diffusion coefficients and thermal
    properties of non-aqueous electrolyte solutions for rechargeable lithium batteries.
    Journal of power sources 81 (1999): 859-862.
    .. [2] John Newman, Dualfoil

    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature


    Returns
    -------
    :class:`pybamm.Symbol`
        Electrolyte diffusivity
    """

    D_c_e = 5.35 * 10 ** (-10)
    E_D_e = 37040
    arrhenius = np.exp(E_D_e / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return D_c_e * arrhenius


def electrolyte_conductivity_PeymanMPM(c_e, T):
    """
    Conductivity of LiPF6 in EC:DMC as a function of ion concentration. The original
    data is from [1]. The fit is from Dualfoil [2].

    References
    ----------
    .. [1] C Capiglia et al. 7Li and 19F diffusion coefficients and thermal
    properties of non-aqueous electrolyte solutions for rechargeable lithium batteries.
    Journal of power sources 81 (1999): 859-862.
    .. [2] John Newman, Dualfoil
    Parameters
    ----------
    c_e: :class:`pybamm.Symbol`
        Dimensional electrolyte concentration
    T: :class:`pybamm.Symbol`
        Dimensional temperature


    Returns
    -------
    :class:`pybamm.Symbol`
        Electrolyte conductivity
    """

    sigma_e = 1.3
    E_k_e = 34700
    arrhenius = np.exp(E_k_e / pybamm.constants.R * (1 / 298.15 - 1 / T))

    return sigma_e * arrhenius


# Call dict via a function to avoid errors when editing in place
def get_parameter_values():
    """
    Parameters for a graphite/NMC532 pouch cell from the paper :footcite:t:`Mohtat2020`
    and references therein.

    SEI parameters are example parameters for SEI growth from the papers
    :footcite:t:`Ramadass2004`, :footcite:t:`Ploehn2004`,
    :footcite:t:`Single2018`, :footcite:t:`Safari2008`, and
    :footcite:t:`Yang2017`

    SEI parameters
    ^^^^^^^^^^^^^^

    Parameters for lithium plating are from the paper :footcite:t:`Yang2017`

    .. note::
        This parameter set does not claim to be representative of the true parameter
        values. Instead these are parameter values that were used to fit SEI models to
        observed experimental data in the referenced papers.
    """

    return {
        "chemistry": "lithium_ion",
        # lithium plating
        "Lithium metal partial molar volume [m3.mol-1]": 1.3e-05,
        "Exchange-current density for plating [A.m-2]": 0.001,
        "Initial plated lithium concentration [mol.m-3]": 0.0,
        "Typical plated lithium concentration [mol.m-3]": 1000.0,
        "Lithium plating transfer coefficient": 0.7,
        # sei
        "Ratio of lithium moles to SEI moles": 2.0,
        "SEI partial molar volume [m3.mol-1]": 9.585e-05,
        "SEI reaction exchange current density [A.m-2]": 1.5e-07,
        "SEI resistivity [Ohm.m]": 200000.0,
        "SEI solvent diffusivity [m2.s-1]": 2.5e-22,
        "Bulk solvent concentration [mol.m-3]": 2636.0,
        "SEI open-circuit potential [V]": 0.4,
        "SEI electron conductivity [S.m-1]": 8.95e-14,
        "SEI lithium interstitial diffusivity [m2.s-1]": 1e-20,
        "Lithium interstitial reference concentration [mol.m-3]": 15.0,
        "Initial SEI thickness [m]": 5e-09,
        "EC initial concentration in electrolyte [mol.m-3]": 4541.0,
        "EC diffusivity [m2.s-1]": 2e-18,
        "SEI kinetic rate constant [m.s-1]": 1e-12,
        "SEI growth activation energy [J.mol-1]": 0.0,
        "Negative electrode reaction-driven LAM factor [m3.mol-1]": 0.0,
        "Positive electrode reaction-driven LAM factor [m3.mol-1]": 0.0,
        # cell
        "Negative current collector thickness [m]": 2.5e-05,
        "Negative electrode thickness [m]": 6.2e-05,
        "Separator thickness [m]": 1.2e-05,
        "Positive electrode thickness [m]": 6.7e-05,
        "Positive current collector thickness [m]": 2.5e-05,
        "Electrode height [m]": 1.0,
        "Electrode width [m]": 0.205,
        "Cell cooling surface area [m2]": 0.41,
        "Cell volume [m3]": 3.92e-05,
        "Negative current collector conductivity [S.m-1]": 59600000.0,
        "Positive current collector conductivity [S.m-1]": 35500000.0,
        "Negative current collector density [kg.m-3]": 8954.0,
        "Positive current collector density [kg.m-3]": 2707.0,
        "Negative current collector specific heat capacity [J.kg-1.K-1]": 385.0,
        "Positive current collector specific heat capacity [J.kg-1.K-1]": 897.0,
        "Negative current collector thermal conductivity [W.m-1.K-1]": 401.0,
        "Positive current collector thermal conductivity [W.m-1.K-1]": 237.0,
        "Nominal cell capacity [A.h]": 5.0,
        "Current function [A]": 5.0,
        "Contact resistance [Ohm]": 0,
        # negative electrode
        "Negative electrode conductivity [S.m-1]": 100.0,
        "Maximum concentration in negative electrode [mol.m-3]": 28746.0,
        "Negative particle diffusivity [m2.s-1]": graphite_diffusivity_PeymanMPM,
        "Negative electrode OCP [V]": graphite_ocp_PeymanMPM,
        "Negative electrode porosity": 0.3,
        "Negative electrode active material volume fraction": 0.61,
        "Negative particle radius [m]": 2.5e-06,
        "Negative electrode Bruggeman coefficient (electrode)": 1.5,
        "Negative electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Negative electrode transport efficiency": 0.16,
        "Negative electrode reference exchange-current density [A.m-2(m3.mol)1.5]"
        "": 1.061e-06,
        "Negative electrode charge transfer coefficient": 0.5,
        "Negative electrode double-layer capacity [F.m-2]": 0.2,
        "Negative electrode exchange-current density [A.m-2]"
        "": graphite_electrolyte_exchange_current_density_PeymanMPM,
        "Negative electrode density [kg.m-3]": 3100.0,
        "Negative electrode specific heat capacity [J.kg-1.K-1]": 1100.0,
        "Negative electrode thermal conductivity [W.m-1.K-1]": 1.7,
        "Negative electrode OCP entropic change [V.K-1]"
        "": graphite_entropic_change_PeymanMPM,
        # positive electrode
        "Positive electrode conductivity [S.m-1]": 100.0,
        "Maximum concentration in positive electrode [mol.m-3]": 35380.0,
        "Positive particle diffusivity [m2.s-1]": NMC_diffusivity_PeymanMPM,
        "Positive electrode OCP [V]": NMC_ocp_PeymanMPM,
        "Positive electrode porosity": 0.3,
        "Positive electrode active material volume fraction": 0.445,
        "Positive particle radius [m]": 3.5e-06,
        "Positive electrode Bruggeman coefficient (electrode)": 1.5,
        "Positive electrode Bruggeman coefficient (electrolyte)": 1.5,
        "Positive electrode transport efficiency": 0.16,
        "Positive electrode reference exchange-current density [A.m-2(m3.mol)1.5]"
        "": 4.824e-06,
        "Positive electrode charge transfer coefficient": 0.5,
        "Positive electrode double-layer capacity [F.m-2]": 0.2,
        "Positive electrode exchange-current density [A.m-2]"
        "": NMC_electrolyte_exchange_current_density_PeymanMPM,
        "Positive electrode density [kg.m-3]": 3100.0,
        "Positive electrode specific heat capacity [J.kg-1.K-1]": 1100.0,
        "Positive electrode thermal conductivity [W.m-1.K-1]": 2.1,
        "Positive electrode OCP entropic change [V.K-1]": NMC_entropic_change_PeymanMPM,
        # separator
        "Separator porosity": 0.4,
        "Separator Bruggeman coefficient (electrolyte)": 1.5,
        "Separator density [kg.m-3]": 397.0,
        "Separator specific heat capacity [J.kg-1.K-1]": 700.0,
        "Separator thermal conductivity [W.m-1.K-1]": 0.16,
        "Separator transport efficiency ": 0.25,
        # electrolyte
        "Initial concentration in electrolyte [mol.m-3]": 1000.0,
        "Cation transference number": 0.38,
        "Thermodynamic factor": 1.0,
        "Typical lithium ion diffusivity [m2.s-1]": 5.34e-10,
        "Electrolyte diffusivity [m2.s-1]": electrolyte_diffusivity_PeymanMPM,
        "Electrolyte conductivity [S.m-1]": electrolyte_conductivity_PeymanMPM,
        # experiment
        "Reference temperature [K]": 298.15,
        "Negative current collector surface heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Positive current collector surface heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Negative tab heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Positive tab heat transfer coefficient [W.m-2.K-1]": 0.0,
        "Edge heat transfer coefficient [W.m-2.K-1]": 5.0,
        "Total heat transfer coefficient [W.m-2.K-1]": 5.0,
        "Ambient temperature [K]": 298.15,
        "Number of electrodes connected in parallel to make a cell": 1.0,
        "Number of cells connected in series to make a battery": 1.0,
        "Lower voltage cut-off [V]": 2.8,
        "Upper voltage cut-off [V]": 4.2,
        "Open-circuit voltage at 0% SOC [V]": 2.8,
        "Open-circuit voltage at 100% SOC [V]": 4.2,
        "Initial concentration in negative electrode [mol.m-3]": 48.8682,
        "Initial concentration in positive electrode [mol.m-3]": 31513.0,
        "Initial temperature [K]": 298.15,
        # citations
        "citations": ["Mohtat2020"],
    }



from .base_sei import BaseModel
from .total_sei import TotalSEI
from .sei_thickness import SEIThickness
from .no_sei import NoSEI
from .constant_sei import ConstantSEI
from .sei_growth import SEIGrowth

__all__ = [
    'base_sei', 'constant_sei', 'no_sei', 'sei_growth', 'sei_thickness', 'total_sei'
]



#
# Class for constant SEI thickness
#
import pybamm

from .base_sei import BaseModel


class ConstantSEI(BaseModel):
    """
    Class for SEI with constant concentration.

    Note that there is no SEI current, so we don't need to update the "sum of
    interfacial current densities" variables from
    :class:`pybamm.interface.BaseInterface`

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, options, phase="primary"):
        super().__init__(param, domain, options=options, phase=phase)
        if self.options.electrode_types[domain] == "planar":
            self.reaction_loc = "interface"
        else:
            self.reaction_loc = "full electrode"

    def get_fundamental_variables(self):
        domain = self.domain.lower()
        L_sei_0 = self.phase_param.L_sei_0
        V_bar_sei = self.phase_param.V_bar_sei
        # Constant concentration
        if self.reaction_loc == "interface":
            # c_sei is an interfacial quantity [mol.m-2]
            c_sei = L_sei_0 / V_bar_sei
        else:
            # c_sei is a bulk quantity [mol.m-3]
            c_sei = L_sei_0 * self.phase_param.a_typ / V_bar_sei
        variables = self._get_standard_concentration_variables(c_sei)

        # Reactions
        if self.reaction_loc == "interface":
            zero = pybamm.PrimaryBroadcast(pybamm.Scalar(0), "current collector")
        else:
            zero = pybamm.FullBroadcast(
                pybamm.Scalar(0), f"{domain} electrode", "current collector"
            )
        variables.update(self._get_standard_reaction_variables(zero))

        return variables



#
# Total interfacial current class, summing up contributions from all reactions
#
import pybamm


class TotalInterfacialCurrent(pybamm.BaseSubModel):
    """
    Total interfacial current, summing up contributions from all reactions

    Parameters
    ----------
    param :
        model parameters
    chemistry : str
        The name of the battery chemistry whose reactions need to be summed up
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    """

    def __init__(self, param, chemistry, options):
        super().__init__(param, options=options)
        self.chemistry = chemistry

    def get_coupled_variables(self, variables):
        """
        Get variables associated with interfacial current over the whole cell domain
        This function also creates the "total source term" variables by summing all
        the reactions.
        """
        for domain in self.options.whole_cell_domains:
            if domain != "separator":
                variables.update(
                    self._get_coupled_variables_by_domain(variables, domain.split()[0])
                )

        variables.update(self._get_whole_cell_coupled_variables(variables))

        return variables

    def _get_coupled_variables_by_domain(self, variables, domain):
        phase_names = [""]

        num_phases = int(getattr(self.options, domain)["particle phases"])
        if num_phases > 1:
            phase_names += ["primary ", "secondary "]

        for phase_name in phase_names:
            variables.update(
                self._get_coupled_variables_by_phase_and_domain(
                    variables, domain, phase_name
                )
            )

        return variables

    def _get_coupled_variables_by_phase_and_domain(self, variables, domain, phase_name):
        Domain = domain.capitalize()

        if self.chemistry == "lithium-ion":
            reaction_names = [""]
            if phase_name == "":
                reaction_names += ["SEI "]
                if self.options.electrode_types[domain] == "porous":
                    # separate plating reaction only if the electrode is porous,
                    # since plating is the main reaction otherwise.
                    # Likewise, SEI on cracks only in a porous electrode
                    reaction_names.extend(["lithium plating ", "SEI on cracks "])
        elif self.chemistry == "lead-acid":
            reaction_names = ["", "oxygen "]

        # Create separate 'new_variables' so that variables only get updated once
        # everything is computed
        new_variables = variables.copy()

        # Initialize "total reaction" variables
        # These will get populated by each reaction, and then used
        # later by "set_rhs" or "set_algebraic", which ensures that we always have
        # added all the necessary variables by the time the sum is used
        new_variables.update(
            {
                f"Sum of {domain} electrode {phase_name}"
                "electrolyte reaction source terms [A.m-3]": 0,
                f"Sum of x-averaged {domain} electrode {phase_name}"
                "electrolyte reaction source terms [A.m-3]": 0,
                f"Sum of {domain} electrode {phase_name}"
                "volumetric interfacial current densities [A.m-3]": 0,
                f"Sum of x-averaged {domain} electrode {phase_name}"
                "volumetric interfacial current densities [A.m-3]": 0,
            }
        )
        for reaction_name in reaction_names:
            # Sum variables
            a_j_k = new_variables[
                f"{Domain} electrode {reaction_name}{phase_name}volumetric "
                "interfacial current density [A.m-3]"
            ]

            if self.chemistry == "lithium-ion":
                # Both the main reaction current contribute to the electrolyte reaction
                # current
                s_k = 1
            elif self.chemistry == "lead-acid":
                if reaction_name == "":  # main reaction
                    s_k = self.param.domain_params[domain].prim.s_plus_S
                elif reaction_name == "oxygen ":
                    s_k = self.param.s_plus_Ox

            new_variables[
                f"Sum of {domain} electrode {phase_name}"
                "electrolyte reaction source terms [A.m-3]"
            ] += s_k * a_j_k
            new_variables[
                f"Sum of x-averaged {domain} electrode {phase_name}"
                "electrolyte reaction source terms [A.m-3]"
            ] += pybamm.x_average(s_k * a_j_k)

            new_variables[
                f"Sum of {domain} electrode {phase_name}volumetric "
                "interfacial current densities [A.m-3]"
            ] += a_j_k
            new_variables[
                f"Sum of x-averaged {domain} electrode {phase_name}"
                "volumetric interfacial current densities [A.m-3]"
            ] += pybamm.x_average(a_j_k)

        variables.update(new_variables)

        return variables

    def _get_whole_cell_coupled_variables(self, variables):
        # Interfacial current density and exchange-current density for the main reaction
        zero_s = pybamm.FullBroadcast(0, "separator", "current collector")

        if (
            self.options.negative["particle phases"] == "1"
            and self.options.positive["particle phases"] == "1"
        ):
            for variable_template in [
                "{}interfacial current density [A.m-2]",
                "{}exchange current density [A.m-2]",
            ]:
                var_dict = {}
                for domain in self.options.whole_cell_domains:
                    if domain == "separator":
                        var_dict[domain] = zero_s
                    else:
                        Domain = domain.capitalize()
                        var_dict[domain] = variables[
                            variable_template.format(Domain + " ")
                        ]
                var = pybamm.concatenation(*var_dict.values())
                var_name = variable_template.format("")
                var_name = var_name[0].upper() + var_name[1:]  # capitalise first letter
                variables.update({var_name: var})

        # Sum variables
        for variable_template in [
            "Sum of {}volumetric interfacial current densities [A.m-3]",
            "Sum of {}electrolyte reaction source terms [A.m-3]",
        ]:
            var_dict = {}
            for domain in self.options.whole_cell_domains:
                if domain == "separator":
                    var_dict[domain] = zero_s
                else:
                    var = variables[variable_template.format(domain + " ")]
                    var_dict[domain] = var
            var = pybamm.concatenation(*var_dict.values())
            variables.update({variable_template.format(""): var})

        return variables



import numpy as np
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import inv

import pybamm
from pybamm.util import import_optional_dependency


class ScikitFiniteElement3D(pybamm.SpatialMethod):
    """
    A class which implements the steps specific to the 3D finite element method
    during discretisation using scikit-fem.

    Parameters
    ----------
    options : dict-like, optional
        A dictionary of options to be passed to the spatial method.
    """

    def __init__(self, options=None):
        super().__init__(options)
        pybamm.citations.register("Gustafsson2020")

    def build(self, mesh):
        super().build(mesh)
        for dom in mesh.keys():
            mesh[dom].npts_for_broadcast_to_nodes = mesh[dom].npts

    def spatial_variable(self, symbol):
        """
        Creates a discretised spatial variable for x, y, z, r or theta.

        Parameters
        ----------
        symbol : :class:`pybamm.SpatialVariable`
            The spatial variable to discretise (must be 'x', 'y', 'z', 'r' or 'theta')

        Returns
        -------
        :class:`pybamm.Vector`
            The discretised spatial variable as a vector of nodal values
        """
        domain = symbol.domain[0]
        mesh = self.mesh[domain]

        if symbol.name == "x":
            entries = mesh.nodes[:, 0][:, np.newaxis]
        elif symbol.name == "y":
            entries = mesh.nodes[:, 1][:, np.newaxis]
        elif symbol.name == "z":
            entries = mesh.nodes[:, 2][:, np.newaxis]
        elif symbol.name == "r" or symbol.name == "r_macro":
            entries = np.sqrt(mesh.nodes[:, 0] ** 2 + mesh.nodes[:, 1] ** 2)[
                :, np.newaxis
            ]
        elif symbol.name == "theta":
            entries = np.arctan2(mesh.nodes[:, 1], mesh.nodes[:, 0])[
                :, np.newaxis
            ]  # pragma: no cover
        else:
            raise pybamm.GeometryError(
                f"Spatial variable must be 'x', 'y', 'z', 'r' or 'theta', not {symbol.name}"
            )  # pragma: no cover
        return pybamm.Vector(entries, domains=symbol.domains)

    def gradient(self, symbol, discretised_symbol, boundary_conditions):
        """
        Matrix-vector multiplication to implement the 3D gradient operator.
        Returns a Concatenation of [grad_x, grad_y, grad_z] or [grad_r, grad_theta, grad_z].

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol that we will take the gradient of.
        discretised_symbol: :class:`pybamm.Symbol`
            The discretised symbol of the correct size
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Concatenation`
            The 3D gradient as concatenation of x, y, z or z, r, theta components
        """

        skfem = import_optional_dependency("skfem")
        domain = symbol.domain[0]
        mesh = self.mesh[domain]
        coord_sys = mesh.coord_sys

        grad_x_matrix, grad_y_matrix, grad_z_matrix = self.gradient_matrix(
            symbol, boundary_conditions
        )

        @skfem.BilinearForm
        def mass_form(u, v, w):
            return u * v

        mass = skfem.asm(mass_form, mesh.basis)
        mass_inv = pybamm.Matrix(inv(csc_matrix(mass)))

        if isinstance(discretised_symbol, pybamm.Scalar):  # pragma: no cover
            zeros = pybamm.Vector(np.zeros((mesh.npts, 1)))
            grad = pybamm.Concatenation(zeros, zeros, zeros, check_domain=False)
            grad.copy_domains(symbol)
            return grad

        grad_x = mass_inv @ (grad_x_matrix @ discretised_symbol)
        grad_y = mass_inv @ (grad_y_matrix @ discretised_symbol)
        grad_z = mass_inv @ (grad_z_matrix @ discretised_symbol)

        if coord_sys == "cartesian":
            grad = pybamm.Concatenation(
                grad_x, grad_y, grad_z, check_domain=False, concat_fun=np.hstack
            )
        elif coord_sys == "cylindrical polar":
            x_nodes = pybamm.Vector(mesh.nodes[:, 0][:, np.newaxis])
            y_nodes = pybamm.Vector(mesh.nodes[:, 1][:, np.newaxis])
            r_nodes = pybamm.sqrt(x_nodes**2 + y_nodes**2)
            r_nodes.entries[r_nodes.entries < 1e-12] = 1e-12

            cos_theta = x_nodes / r_nodes
            sin_theta = y_nodes / r_nodes

            grad_r = grad_x * cos_theta + grad_y * sin_theta
            grad_theta = -grad_x * sin_theta + grad_y * cos_theta

            grad = pybamm.Concatenation(
                grad_r, grad_theta, grad_z, check_domain=False, concat_fun=np.hstack
            )
        else:
            raise pybamm.DiscretisationError(
                f"Unknown coordinate system '{coord_sys}'"
            )  # pragma: no cover

        grad.copy_domains(symbol)

        return grad

    def gradient_squared(self, symbol, discretised_symbol, boundary_conditions):
        """
        Multiplication to implement the inner product of the gradient operator
        with itself in 3D.
        """
        grad = self.gradient(symbol, discretised_symbol, boundary_conditions)
        grad_x, grad_y, grad_z = grad.orphans
        return grad_x**2 + grad_y**2 + grad_z**2

    def gradient_matrix(self, symbol, boundary_conditions):
        """
        Gradient matrices for finite elements in 3D.

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol for which we want to calculate the gradient matrix
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        tuple of :class:`pybamm.Matrix`
            The (sparse) finite element gradient matrices.
        """
        skfem = import_optional_dependency("skfem")
        domain = symbol.domain[0]
        mesh = self.mesh[domain]

        @skfem.BilinearForm
        def gradient_dx(u, v, w):
            return u.grad[0] * v

        @skfem.BilinearForm
        def gradient_dy(u, v, w):
            return u.grad[1] * v

        @skfem.BilinearForm
        def gradient_dz(u, v, w):
            return u.grad[2] * v

        grad_x = skfem.asm(gradient_dx, mesh.basis)
        grad_y = skfem.asm(gradient_dy, mesh.basis)
        grad_z = skfem.asm(gradient_dz, mesh.basis)

        return pybamm.Matrix(grad_x), pybamm.Matrix(grad_y), pybamm.Matrix(grad_z)

    def laplacian(self, symbol, discretised_symbol, boundary_conditions):
        """
        Matrix-vector multiplication to implement the 3D Laplacian operator.
        This should be called only after boundary conditions have been properly
        discretised by the PyBaMM discretisation process.

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol for which we want to calculate the Laplacian
        discretised_symbol: :class:`pybamm.Symbol`
            The discretised symbol of the correct size
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Symbol`
            The Laplacian of the symbol
        """
        stiffness_matrix = self.stiffness_matrix(symbol, boundary_conditions)
        boundary_load = self.laplacian_boundary_load(symbol, boundary_conditions)

        return -stiffness_matrix @ discretised_symbol + boundary_load

    def divergence(self, symbol, discretised_symbol, boundary_conditions):
        """
        Matrix-vector multiplication to implement the 3D divergence operator.
        Expects discretised_symbol to be a Concatenation of [Fx, Fy, Fz].

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol representing the divergence operation
        discretised_symbol: :class:`pybamm.Concatenation`
            The discretised vector field as concatenation [Fx, Fy, Fz]
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Symbol`
            The divergence of the vector field
        """
        domain_key = symbol.domain[0]

        if not (
            hasattr(discretised_symbol, "orphans")
            and len(discretised_symbol.orphans) == 3
        ):
            raise ValueError(
                "divergence expects a concatenation of 3 vector components"
            )  # pragma: no cover

        Fx_var = pybamm.Variable("Fx", domain=[domain_key])
        grad_M = self.gradient_matrix(Fx_var, boundary_conditions)

        skfem = import_optional_dependency("skfem")
        mesh = self.mesh[domain_key]
        coord_sys = mesh.coord_sys

        if coord_sys == "cartesian":
            Fx, Fy, Fz = discretised_symbol.orphans
        elif coord_sys == "cylindrical polar":
            Fr, Ftheta, Fz_cyl = discretised_symbol.orphans
            x_nodes = pybamm.Vector(mesh.nodes[:, 0][:, np.newaxis])
            y_nodes = pybamm.Vector(mesh.nodes[:, 1][:, np.newaxis])
            r_nodes = pybamm.sqrt(x_nodes**2 + y_nodes**2)
            r_nodes.entries[r_nodes.entries < 1e-12] = 1e-12

            cos_theta = x_nodes / r_nodes
            sin_theta = y_nodes / r_nodes

            Fx = Fr * cos_theta - Ftheta * sin_theta
            Fy = Fr * sin_theta + Ftheta * cos_theta
            Fz = Fz_cyl
        else:
            raise pybamm.DiscretisationError(
                f"Unknown coordinate system '{coord_sys}'"
            )  # pragma: no cover

        @skfem.BilinearForm
        def mass_form(u, v, w):
            return u * v

        mass = skfem.asm(mass_form, mesh.basis)
        mass_inv = pybamm.Matrix(inv(csc_matrix(mass)))

        grad_x_T = pybamm.Matrix(grad_M[0].entries)
        grad_y_T = pybamm.Matrix(grad_M[1].entries)
        grad_z_T = pybamm.Matrix(grad_M[2].entries)

        div_x = mass_inv @ (grad_x_T @ Fx)
        div_y = mass_inv @ (grad_y_T @ Fy)
        div_z = mass_inv @ (grad_z_T @ Fz)

        div_result = div_x + div_y + div_z
        div_result.clear_domains()
        return div_result

    def stiffness_matrix(self, symbol, boundary_conditions):
        """
        Laplacian (stiffness) matrix for finite elements in 3D.

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol for which we want to calculate the stiffness matrix
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) finite element stiffness matrix
        """
        skfem = import_optional_dependency("skfem")
        domain = symbol.domain[0]
        mesh = self.mesh[domain]

        @skfem.BilinearForm
        def stiffness_form(u, v, w):
            return u.grad[0] * v.grad[0] + u.grad[1] * v.grad[1] + u.grad[2] * v.grad[2]

        stiffness = skfem.asm(stiffness_form, mesh.basis)

        if symbol in boundary_conditions:
            bcs = boundary_conditions[symbol]
            for name, (_, bc_type) in bcs.items():
                if bc_type == "Dirichlet":
                    boundary_dofs = getattr(mesh, f"{name}_dofs", None)
                    if boundary_dofs is not None and len(boundary_dofs) > 0:
                        self.bc_apply(stiffness, boundary_dofs)

        return pybamm.Matrix(stiffness)

    def laplacian_boundary_load(self, symbol_for_laplacian, boundary_conditions_dict):
        """
        Assembles the boundary load vector for the 3D Laplacian.

        Parameters
        ----------
        symbol: :class:`pybamm.Symbol`
            The symbol for which we want to calculate the boundary load
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Vector`
            The boundary load vector
        """
        skfem = import_optional_dependency("skfem")
        domain = symbol_for_laplacian.domain[0]
        fem_mesh = self.mesh[domain]

        current_boundary_load_symbol = pybamm.Vector(
            np.zeros((fem_mesh.npts, 1)), domains=symbol_for_laplacian.domains
        )

        bcs_for_symbol = boundary_conditions_dict.get(symbol_for_laplacian, {})

        unit_bc_load_form = None
        if any(bc_type == "Neumann" for _, (_, bc_type) in bcs_for_symbol.items()):

            @skfem.LinearForm
            def _unit_bc_load_form(v, w):
                return v

            unit_bc_load_form = _unit_bc_load_form
        if symbol_for_laplacian in boundary_conditions_dict:
            for name, (bc_value_symbol, bc_type) in bcs_for_symbol.items():
                term_contribution = None
                if bc_type == "Neumann":
                    numeric_coeffs = skfem.asm(
                        unit_bc_load_form, getattr(fem_mesh, f"{name}_basis")
                    )
                    if numeric_coeffs.ndim == 1:
                        numeric_coeffs = numeric_coeffs[:, np.newaxis]
                    term_contribution = bc_value_symbol * pybamm.Vector(numeric_coeffs)
                    current_boundary_load_symbol += term_contribution
                elif bc_type == "Dirichlet":
                    boundary_dofs = getattr(fem_mesh, f"{name}_dofs")
                    numeric_mask = np.zeros((fem_mesh.npts, 1))
                    numeric_mask[boundary_dofs, 0] = 1.0
                    term_contribution = bc_value_symbol * pybamm.Vector(numeric_mask)
                    current_boundary_load_symbol += term_contribution

        return current_boundary_load_symbol

    def integral(
        self, child, discretised_child, integration_dimension, integration_variable
    ):
        """
        Vector-vector dot product to implement the 3D integral operator.

        Parameters
        ----------
        child : :class:`pybamm.Symbol`
            The symbol being integrated
        discretised_child : :class:`pybamm.Symbol`
            The discretised symbol being integrated (vector of nodal values)
        integration_dimension : str
            The dimension over which to integrate (e.g. "primary" for volume)

        Returns
        -------
        :class:`pybamm.Symbol`
            The result of the integration (a scalar value).
        """
        integration_vector = self.definite_integral_matrix(child)
        out = integration_vector @ discretised_child
        return out

    def definite_integral_matrix(self, child, vector_type="row"):
        """
        Matrix for definite integral *vector* (vector of ones).
        The child is used to determine the domain and size.

        Parameters
        ----------
        child : :class:`pybamm.Symbol`
            The symbol whose domain/mesh determines the size.
        vector_type : str, optional
            "row" or "column" for the shape of the resulting vector of ones.

        Returns
        -------
        :class:`pybamm.Matrix`
            A matrix representing a vector of ones (either row or column).
        """
        skfem = import_optional_dependency("skfem")
        # get primary domain mesh
        domain = child.domain[0]
        mesh = self.mesh[domain]

        # make form for the integral
        @skfem.LinearForm
        def integral_form(v, w):
            return v

        vector = skfem.asm(integral_form, mesh.basis)

        if vector_type == "row":
            return pybamm.Matrix(vector[np.newaxis, :])
        elif vector_type == "column":  # pragma: no cover
            return pybamm.Matrix(vector[:, np.newaxis])

    def indefinite_integral(self, child, discretised_child, direction):
        """
        Implementation of the indefinite integral operator in 3D.
        """
        raise NotImplementedError(
            "Indefinite integral not implemented for 3D finite elements"
        )

    def boundary_integral(self, child, discretised_child, region):
        """
        Implementation of the 3D boundary integral operator.

        Parameters
        ----------
        child : :class:`pybamm.Symbol`
            The symbol being integrated
        discretised_child : :class:`pybamm.Symbol`
            The discretised symbol being integrated
        region : str
            The boundary region over which to integrate

        Returns
        -------
        :class:`pybamm.Symbol`
            The result of the boundary integration
        """
        integration_vector = self.boundary_integral_vector(child.domain, region=region)
        out = integration_vector @ discretised_child
        out.clear_domains()
        return out

    def boundary_integral_vector(self, domain, region):
        """
        A vector representing an integral operator over the boundary.

        Parameters
        ----------
        domain : list
            The domain(s) of the variable in the integrand
        region : str
            The region of the boundary over which to integrate

        Returns
        -------
        :class:`pybamm.Matrix`
            The finite element boundary integral vector
        """
        skfem = import_optional_dependency("skfem")
        mesh = self.mesh[domain[0]]

        @skfem.LinearForm
        def integral_form(v, w):
            return v

        if region == "entire":
            integration_vector = skfem.asm(integral_form, mesh.facet_basis)
        elif hasattr(mesh, f"{region}_basis"):
            integration_vector = skfem.asm(
                integral_form, getattr(mesh, f"{region}_basis")
            )

        return pybamm.Matrix(integration_vector[np.newaxis, :])

    def boundary_value_or_flux(self, symbol, discretised_child, bcs=None):
        """
        Returns the boundary value or flux of a variable in 3D.

        Parameters
        ----------
        symbol : :class:`pybamm.Symbol`
            The boundary symbol
        discretised_child : :class:`pybamm.Symbol`
            The discretised variable
        bcs : dict, optional
            Boundary conditions

        Returns
        -------
        :class:`pybamm.Symbol`
            The boundary value or flux
        """
        domain = symbol.children[0].domain
        region = symbol.side

        if isinstance(symbol, pybamm.BoundaryValue):
            integration_vector = self.boundary_integral_vector(domain, region=region)
            boundary_val_vector = integration_vector / (
                integration_vector @ pybamm.Vector(np.ones(integration_vector.shape[1]))
            )
            boundary_value = boundary_val_vector @ discretised_child
            boundary_value.copy_domains(symbol)
            return boundary_value

        elif isinstance(symbol, pybamm.BoundaryGradient):
            raise NotImplementedError(
                "Boundary gradient not implemented for 3D finite elements"
            )  # pragma: no cover

    def mass_matrix(self, symbol, boundary_conditions):
        """
        Calculates the mass matrix for the 3D finite element method.

        Parameters
        ----------
        symbol: :class:`pybamm.Variable`
            The variable corresponding to the equation
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) mass matrix for the spatial method.
        """
        return self.assemble_mass_form(symbol, boundary_conditions)

    def boundary_mass_matrix(self, symbol, boundary_conditions):
        """
        Calculates the mass matrix assembled over the 3D boundary.

        Parameters
        ----------
        symbol: :class:`pybamm.Variable`
            The variable corresponding to the equation
        boundary_conditions : dict
            The boundary conditions of the model

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) boundary mass matrix
        """
        return self.assemble_mass_form(symbol, boundary_conditions, region="boundary")

    def assemble_mass_form(self, symbol, boundary_conditions, region="interior"):
        """
        Assembles the form of the 3D finite element mass matrix.

        Parameters
        ----------
        symbol: :class:`pybamm.Variable`
            The variable corresponding to the equation
        boundary_conditions : dict
            The boundary conditions of the model
        region: str, optional
            The domain over which to assemble

        Returns
        -------
        :class:`pybamm.Matrix`
            The (sparse) mass matrix
        """
        skfem = import_optional_dependency("skfem")
        domain = symbol.domain[0]
        mesh = self.mesh[domain]

        @skfem.BilinearForm
        def mass_form(u, v, w):
            return u * v

        if region == "interior":
            mass = skfem.asm(mass_form, mesh.basis)
        elif region == "boundary":  # pragma: no cover
            mass = skfem.asm(mass_form, mesh.facet_basis)

        if symbol in boundary_conditions:
            bcs = boundary_conditions[symbol]
            for name, (_, bc_type) in bcs.items():
                if bc_type == "Dirichlet":
                    boundary_dofs = getattr(mesh, f"{name}_dofs", None)
                    if boundary_dofs is not None and len(boundary_dofs) > 0:
                        self.bc_apply(mass, boundary_dofs, zero=True)
        return pybamm.Matrix(mass)

    def bc_apply(self, M, boundary_dofs, zero=False):
        """
        Adjusts matrices for 3D boundary conditions.

        Parameters
        ----------
        M : scipy.sparse matrix
            Matrix to modify
        boundary_dofs : array_like
            Boundary degrees of freedom
        zero : bool, optional
            Whether to zero the diagonal entries
        """
        M_lil = M.tolil()

        for dof in boundary_dofs:
            M_lil[dof, :] = 0
            if not zero:
                M_lil[dof, dof] = 1.0

        M_csr = M_lil.tocsr()
        M.data = M_csr.data
        M.indices = M_csr.indices
        M.indptr = M_csr.indptr
        M._shape = M_csr._shape

    def process_binary_operators(self, bin_op, left, right, disc_left, disc_right):
        """
        Process binary operators, with a specific fix for Inner products
        involving Concatenations that may have been created without a
        concatenation_function.

        Parameters
        ----------
        bin_op : :class:`pybamm.BinaryOperator`
            The binary operator to process.
        left : :class:`pybamm.Symbol`
            The left child of the operator.
        right : :class:`pybamm.Symbol`
            The right child of the operator.
        disc_left : :class:`pybamm.Symbol`
            The discretised left child of the operator.
        disc_right : :class:`pybamm.Symbol`
            The discretised right child of the operator.

        Returns
        -------
        :class:`pybamm.Symbol`
            The discretised binary operator.
        """
        if isinstance(bin_op, pybamm.Inner):
            if (
                isinstance(disc_left, pybamm.Concatenation)
                and disc_left.concatenation_function is None
            ):
                target_domain = bin_op.domain
                new_children = [
                    pybamm.PrimaryBroadcast(child, target_domain)
                    for child in disc_left.children
                ]
                disc_left = pybamm.Concatenation(
                    *new_children, concat_fun=np.hstack, check_domain=False
                )
            if (
                isinstance(disc_right, pybamm.Concatenation)
                and disc_right.concatenation_function is None
            ):
                target_domain = bin_op.domain
                new_children = [
                    pybamm.PrimaryBroadcast(child, target_domain)
                    for child in disc_right.children
                ]
                disc_right = pybamm.Concatenation(
                    *new_children, concat_fun=np.hstack, check_domain=False
                )
        return super().process_binary_operators(
            bin_op, left, right, disc_left, disc_right
        )



from .base_electrolyte_conductivity import BaseElectrolyteConductivity
from .leading_order_conductivity import LeadingOrder
from .composite_conductivity import Composite
from .full_conductivity import Full
from .integrated_conductivity import Integrated

from . import surface_potential_form

__all__ = ['base_electrolyte_conductivity', 'composite_conductivity',
           'full_conductivity', 'integrated_conductivity',
           'leading_order_conductivity', 'surface_potential_form']



import numpy as np

import pybamm
from pybamm.util import import_optional_dependency


def plot_3d_cross_section(
    solution: "pybamm.Solution",
    variable: str,
    t: float | None,
    plane: str = "yz",
    position: float = 0.5,
    n_pts: int = 100,
    ax=None,
    show_plot: bool = True,
    cmap: str = "inferno",
    levels: int = 20,
    use_offset: bool = False,
    show_mesh: bool = False,
    mesh_color: str = "white",
    mesh_alpha: float = 0.4,
    mesh_linewidth: float = 0.7,
    mesh_tolerance: float | None = None,
    **kwargs,
):
    """
    Plots a high-quality 2D cross-section of a 3D variable from a PyBaMM solution,
    with mesh overlay support.

    Parameters
    ----------
    solution : pybamm.Solution
        The solution object containing the 3D variable.
    variable : str, optional
        The name of the 3D variable to plot (default: "Cell temperature [K]").
    t : float, optional
        The time at which to plot. If None, the last timestep is used.
    plane : str, optional
        The plane for the cross-section ('xy', 'yz', 'xz', 'rz').
    position : float, optional
        The relative position (0 to 1) along the third axis to take the slice.
    n_pts : int, optional
        The number of points for the plotting grid in each direction.
    ax : matplotlib.axes.Axes, optional
        The axes on which to draw the plot. If None, a new figure is created.
    show_plot : bool, optional
        Whether to display the plot.
    cmap : str, optional
        The colormap for the plot.
    levels : int, optional
        The number of contour levels for the plot.
    use_offset : bool, optional
        Parameter to control color bar format to use scientific notation (default: False).
    show_mesh : bool, optional
        Whether to overlay the calculated FEM mesh slice on the plot.
    mesh_color : str, optional
        Color of the mesh lines.
    mesh_alpha : float, optional
        Transparency of the mesh lines.
    mesh_linewidth : float, optional
        Width of the mesh lines.
    **kwargs
        Additional keyword arguments passed to matplotlib.contourf.
    """
    plt = import_optional_dependency("matplotlib.pyplot")
    model = solution.all_models[0]
    if model.options.get("dimensionality") != 3:
        raise TypeError("This function requires a 3D model solution.")

    if t is None:
        t = solution.t[-1]

    var_obj = solution[variable]
    mesh = var_obj.mesh
    nodes = mesh.nodes
    elements = mesh.elements

    x_min, x_max = np.min(nodes[:, 0]), np.max(nodes[:, 0])
    y_min, y_max = np.min(nodes[:, 1]), np.max(nodes[:, 1])
    z_min, z_max = np.min(nodes[:, 2]), np.max(nodes[:, 2])
    geometry = model.options.get("cell geometry", "pouch")

    if mesh_tolerance is None:
        domain_size = max(x_max - x_min, y_max - y_min, z_max - z_min)
        mesh_tolerance = domain_size * 0.01  # 1% of domain size

    fig = None
    if ax is None:
        if geometry == "cylindrical" and plane == "xy":
            fig, ax = plt.subplots(subplot_kw={"projection": "polar"})
        else:
            fig, ax = plt.subplots(figsize=(8, 6))

    fig = fig or ax.get_figure()
    title_suffix = f"at t={t:.1f}s"

    def add_mesh_overlay(
        ax, nodes, elements, plane, slice_coord_val, geometry, mesh_tolerance
    ):
        """Calculates and draws the true intersection of the mesh with the plane."""

        if plane == "yz":
            slice_axis_idx = 0  # slice on x
            plot_axes_indices = [1, 2]  # plot y,z
        elif plane == "xz":
            slice_axis_idx = 1  # slice on y
            plot_axes_indices = [0, 2]  # plot x,z
        elif plane == "xy":
            slice_axis_idx = 2  # slice on z
            plot_axes_indices = [0, 1]  # plot x,y
        elif plane == "rz":
            slice_axis_idx = 1  # slice on y
            slice_coord_val = 0.0

        mesh_segments = []

        for element in elements:
            intersection_points = []
            edges = [
                (element[0], element[1]),
                (element[0], element[2]),
                (element[0], element[3]),
                (element[1], element[2]),
                (element[1], element[3]),
                (element[2], element[3]),
            ]

            for p1_idx, p2_idx in edges:
                p1, p2 = nodes[p1_idx], nodes[p2_idx]
                c1, c2 = p1[slice_axis_idx], p2[slice_axis_idx]

                # Check if edge crosses the slice plane
                if abs(c1 - slice_coord_val) <= mesh_tolerance:
                    intersection_points.append(p1)
                elif abs(c2 - slice_coord_val) <= mesh_tolerance:
                    intersection_points.append(p2)
                elif (c1 < slice_coord_val < c2) or (c2 < slice_coord_val < c1):
                    # Edge crosses the plane
                    ratio = (slice_coord_val - c1) / (c2 - c1)
                    intersection_point = p1 + ratio * (p2 - p1)
                    intersection_points.append(intersection_point)

            # Remove duplicate points
            if len(intersection_points) >= 2:
                unique_points = []
                for pt in intersection_points:
                    is_duplicate = False
                    for existing_pt in unique_points:
                        if np.linalg.norm(pt - existing_pt) < mesh_tolerance:
                            is_duplicate = True
                            break
                    if not is_duplicate:
                        unique_points.append(pt)

                if len(unique_points) >= 2:
                    poly_xyz = np.array(unique_points)

                    if geometry == "cylindrical" and plane == "xy":
                        x_coords, y_coords = poly_xyz[:, 0], poly_xyz[:, 1]
                        r_coords = np.sqrt(x_coords**2 + y_coords**2)
                        theta_coords = np.arctan2(y_coords, x_coords)
                        theta_coords[theta_coords < 0] += (
                            2 * np.pi
                        )  # Ensure positive angles
                        plot_coords = np.column_stack([theta_coords, r_coords])
                    elif geometry == "cylindrical" and plane == "rz":
                        mask = poly_xyz[:, 0] >= 0
                        if np.sum(mask) >= 2:
                            filtered_points = poly_xyz[mask]
                            r_coords = np.sqrt(
                                filtered_points[:, 0] ** 2 + filtered_points[:, 1] ** 2
                            )
                            z_coords = filtered_points[:, 2]
                            plot_coords = np.column_stack([r_coords, z_coords])
                        else:
                            continue
                    else:  # Cartesian
                        plot_coords = poly_xyz[:, plot_axes_indices]

                    if len(plot_coords) >= 2:
                        mesh_segments.append(plot_coords)

        segments_plotted = 0
        for segment in mesh_segments:
            if len(segment) >= 2:
                if len(segment) > 2:
                    # Sort points to form a proper polygon
                    centroid = np.mean(segment, axis=0)
                    angles = np.arctan2(
                        segment[:, 1] - centroid[1], segment[:, 0] - centroid[0]
                    )
                    sorted_segment = segment[np.argsort(angles)]
                    # Close the polygon
                    final_segment = np.vstack([sorted_segment, sorted_segment[0]])
                else:
                    final_segment = segment

                ax.plot(
                    final_segment[:, 0],
                    final_segment[:, 1],
                    color=mesh_color,
                    alpha=mesh_alpha,
                    linewidth=mesh_linewidth,
                )
                segments_plotted += 1

        print(f"Plotted {segments_plotted} mesh segments")

    x_label, y_label = "", ""

    slice_coord_val = None
    if geometry == "cylindrical":
        r_coords = np.sqrt(nodes[:, 0] ** 2 + nodes[:, 1] ** 2)
        r_min, r_max = np.min(r_coords), np.max(r_coords)

        if plane == "xy":
            slice_coord_val = z_min + (z_max - z_min) * position
            r_grid = np.linspace(r_min, r_max, n_pts)
            theta_grid = np.linspace(0, 2 * np.pi, n_pts)
            R_mesh, Theta_mesh = np.meshgrid(r_grid, theta_grid)
            X_eval, Y_eval = R_mesh * np.cos(Theta_mesh), R_mesh * np.sin(Theta_mesh)
            Z_eval = np.full_like(X_eval, slice_coord_val)
            data = var_obj(t=t, x=X_eval, y=Y_eval, z=Z_eval)
            pcm = ax.contourf(
                Theta_mesh, R_mesh, data, levels=levels, cmap=cmap, **kwargs
            )
            ax.set_ylim(r_min, r_max)
            plot_title = f"T(r,θ) at z={slice_coord_val:.2f}m, {title_suffix}"

        elif plane == "rz":
            slice_coord_val = 0.0  # Slice at theta=0
            r_grid = np.linspace(r_min, r_max, n_pts)
            z_grid = np.linspace(z_min, z_max, n_pts)
            R_mesh, Z_mesh = np.meshgrid(r_grid, z_grid)
            X_eval, Y_eval = R_mesh, np.zeros_like(R_mesh)
            Z_eval = Z_mesh
            data = var_obj(t=t, x=X_eval, y=Y_eval, z=Z_eval)
            pcm = ax.contourf(R_mesh, Z_mesh, data, levels=levels, cmap=cmap, **kwargs)
            x_label, y_label = "Radius r [m]", "Height z [m]"
            plot_title = f"T(r,z) Cross-Section, {title_suffix}"
        else:
            raise ValueError(f"Plane '{plane}' invalid for cylindrical geometry.")

    else:  # Cartesian geometry
        if plane == "yz":
            slice_coord_val = x_min + (x_max - x_min) * position
            grid_1, grid_2 = (
                np.linspace(y_min, y_max, n_pts),
                np.linspace(z_min, z_max, n_pts),
            )
            Y_eval, Z_eval = np.meshgrid(grid_1, grid_2)
            X_eval = np.full_like(Y_eval, slice_coord_val)
            data = var_obj(
                t=t, x=X_eval.ravel(), y=Y_eval.ravel(), z=Z_eval.ravel()
            ).reshape(X_eval.shape)
            pcm = ax.contourf(Y_eval, Z_eval, data, levels=levels, cmap=cmap, **kwargs)
            x_label, y_label = "y [m]", "z [m]"
            plot_title = f"T(y,z) at x={slice_coord_val:.2f}m, {title_suffix}"
        elif plane == "xz":
            slice_coord_val = y_min + (y_max - y_min) * position
            grid_1, grid_2 = (
                np.linspace(x_min, x_max, n_pts),
                np.linspace(z_min, z_max, n_pts),
            )
            X_eval, Z_eval = np.meshgrid(grid_1, grid_2)
            Y_eval = np.full_like(X_eval, slice_coord_val)
            data = var_obj(
                t=t, x=X_eval.ravel(), y=Y_eval.ravel(), z=Z_eval.ravel()
            ).reshape(X_eval.shape)
            pcm = ax.contourf(X_eval, Z_eval, data, levels=levels, cmap=cmap, **kwargs)
            x_label, y_label = "x [m]", "z [m]"
            plot_title = f"T(x,z) at y={slice_coord_val:.2f}m, {title_suffix}"
        elif plane == "xy":
            slice_coord_val = z_min + (z_max - z_min) * position
            grid_1, grid_2 = (
                np.linspace(x_min, x_max, n_pts),
                np.linspace(y_min, y_max, n_pts),
            )
            X_eval, Y_eval = np.meshgrid(grid_1, grid_2)
            Z_eval = np.full_like(X_eval, slice_coord_val)
            data = var_obj(
                t=t, x=X_eval.ravel(), y=Y_eval.ravel(), z=Z_eval.ravel()
            ).reshape(X_eval.shape)
            pcm = ax.contourf(X_eval, Y_eval, data, levels=levels, cmap=cmap, **kwargs)
            x_label, y_label = "x [m]", "y [m]"
            plot_title = f"T(x,y) at z={slice_coord_val:.2f}m, {title_suffix}"
        else:
            raise ValueError(
                f"Plane '{plane}' invalid for Cartesian geometry. Use 'xy', 'yz', or 'xz'."
            )

    # Add mesh overlay
    if show_mesh:
        add_mesh_overlay(
            ax, nodes, elements, plane, slice_coord_val, geometry, mesh_tolerance
        )

    cbar = fig.colorbar(pcm, ax=ax, label=f"{variable}")
    if not use_offset:
        cbar.formatter.set_useOffset(False)

    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_title(plot_title)
    ax.set_aspect("auto", "box")

    if show_plot:
        plt.tight_layout()
        plt.show()

    return ax



__all__ = ['array', 'averages', 'binary_operators', 'broadcasts',
           'concatenations', 'exceptions', 'functions', 'independent_variable',
           'input_parameter', 'interpolant', 'matrix', 'operations',
           'parameter', 'printing', 'scalar', 'state_vector', 'symbol',
           'unary_operators', 'variable', 'vector', 'discrete_time_sum', 'vector_field', 'magnitude' ]



import pybamm
from pybamm.util import import_optional_dependency

from .quick_plot import ax_max, ax_min


def plot2D(x, y, z, ax=None, show_plot=True, **kwargs):
    """
    Generate a simple 2D plot. Calls `matplotlib.pyplot.contourf` with keyword
    arguments 'kwargs'.  For a list of 'kwargs' see the
    `matplotlib contourf documentation
    <https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.contourf.html>`_

    Parameters
    ----------
    x : :class:`pybamm.Array`
        The array to plot on the x axis. Can be of shape (M, N) or (N, 1)
    y : :class:`pybamm.Array`
        The array to plot on the y axis. Can be of shape (M, N)  or (M, 1)
    z : :class:`pybamm.Array`
        The array to plot on the z axis. Is of shape (M, N)
    ax : matplotlib Axis, optional
        The axis on which to put the plot. If None, a new figure and axis is created.
    show_plot : bool, optional
        Whether to show the plots. Default is True. Set to False if you want to
        only display the plot after plt.show() has been called.

    """
    plt = import_optional_dependency("matplotlib.pyplot")

    if not isinstance(x, pybamm.Array):
        raise TypeError("x must be 'pybamm.Array'")
    if not isinstance(y, pybamm.Array):
        raise TypeError("y must be 'pybamm.Array'")
    if not isinstance(z, pybamm.Array):
        raise TypeError("z must be 'pybamm.Array'")

    if ax is not None:
        show_plot = False
    else:
        _, ax = plt.subplots()

    # Get correct entries of x and y depending on shape
    if x.shape == y.shape == z.shape:
        x_entries = x.entries
        y_entries = y.entries
    else:
        x_entries = x.entries[:, 0]
        y_entries = y.entries[:, 0]

    plot = ax.contourf(
        x_entries,
        y_entries,
        z.entries,
        vmin=ax_min(z.entries),
        vmax=ax_max(z.entries),
        **kwargs,
    )
    plt.colorbar(plot, ax=ax)

    if show_plot:  # pragma: no cover
        plt.show()

    return ax



#
# Equivalent Circuit Model with split OCV
#
import pybamm


class SplitOCVR(pybamm.BaseModel):
    """Basic Equivalent Circuit Model that uses two OCV functions
    for each electrode. This model is easily parameterizable with minimal parameters.
    This class differs from the :class: pybamm.equivalent_circuit.Thevenin() due
    to dual OCV functions to make up the voltage from each electrode.

    Parameters
    ----------
    name: str, optional
        The name of the model.
    """

    def __init__(self, name="ECM with split OCV"):
        super().__init__(name)

        ######################
        # Variables
        ######################
        # All variables are only time-dependent
        # No domain definition needed

        theta_n = pybamm.Variable("Negative particle stoichiometry")
        theta_p = pybamm.Variable("Positive particle stoichiometry")
        Q = pybamm.Variable("Discharge capacity [A.h]")
        V = pybamm.Variable("Voltage [V]")

        # model is isothermal
        I = pybamm.FunctionParameter("Current function [A]", {"Time [s]": pybamm.t})

        # Capacity equation
        self.rhs[Q] = I / 3600
        self.initial_conditions[Q] = pybamm.Scalar(0)

        # Capacity in each electrode
        Q_n = pybamm.Parameter("Negative electrode capacity [A.h]")
        Q_p = pybamm.Parameter("Positive electrode capacity [A.h]")

        # State of charge electrode equations
        theta_n_0 = pybamm.Parameter("Negative electrode initial stoichiometry")
        theta_p_0 = pybamm.Parameter("Positive electrode initial stoichiometry")
        self.rhs[theta_n] = -I / Q_n / 3600
        self.rhs[theta_p] = I / Q_p / 3600
        self.initial_conditions[theta_n] = theta_n_0
        self.initial_conditions[theta_p] = theta_p_0

        # Resistance for IR expression
        R = pybamm.Parameter("Ohmic resistance [Ohm]")

        # Open-circuit potential for each electrode
        Un = pybamm.FunctionParameter(
            "Negative electrode OCP [V]", {"Negative particle stoichiometry": theta_n}
        )
        Up = pybamm.FunctionParameter(
            "Positive electrode OCP [V]", {"Positive particle stoichiometry": theta_p}
        )

        # Voltage expression
        V = Up - Un - I * R

        # Parameters for Voltage cutoff
        voltage_high_cut = pybamm.Parameter("Upper voltage cut-off [V]")
        voltage_low_cut = pybamm.Parameter("Lower voltage cut-off [V]")

        self.variables = {
            "Negative particle stoichiometry": theta_n,
            "Positive particle stoichiometry": theta_p,
            "Current [A]": I,
            "Discharge capacity [A.h]": Q,
            "Voltage [V]": V,
            "Times [s]": pybamm.t,
            "Positive electrode OCP [V]": Up,
            "Negative electrode OCP [V]": Un,
            "Current function [A]": I,
        }

        # Events specify points at which a solution should terminate
        self.events += [
            pybamm.Event("Minimum voltage [V]", V - voltage_low_cut),
            pybamm.Event("Maximum voltage [V]", voltage_high_cut - V),
            pybamm.Event("Maximum Negative Electrode stoichiometry", 0.999 - theta_n),
            pybamm.Event("Maximum Positive Electrode stoichiometry", 0.999 - theta_p),
            pybamm.Event("Minimum Negative Electrode stoichiometry", theta_n - 0.0001),
            pybamm.Event("Minimum Positive Electrode stoichiometry", theta_p - 0.0001),
        ]

    @property
    def default_quick_plot_variables(self):
        return [
            "Voltage [V]",
            ["Negative particle stoichiometry", "Positive particle stoichiometry"],
            "Negative electrode OCP [V]",
            "Positive electrode OCP [V]",
            "Current [A]",
        ]



#
# Single open-circuit potential, same for charge and discharge
#
import pybamm

from . import BaseOpenCircuitPotential


class SingleOpenCircuitPotential(BaseOpenCircuitPotential):
    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.reaction == "lithium-ion main":
            sto_surf, sto_bulk, T, T_bulk = self._get_stoichiometry_and_temperature(
                variables
            )

            ocp_surf = self.phase_param.U(sto_surf, T)
            ocp_bulk = self.phase_param.U(sto_bulk, T_bulk)

            dUdT = self.phase_param.dUdT(sto_surf)

            variables.update(
                {
                    f"{Domain} electrode {phase_name}equilibrium open-circuit potential [V]": ocp_surf,
                    f"X-averaged {domain} electrode {phase_name}equilibrium open-circuit potential [V]": pybamm.x_average(
                        ocp_surf
                    ),
                }
            )

        elif self.reaction == "lithium metal plating":
            T = variables[f"{Domain} electrode temperature [K]"]
            ocp_surf = 0 * T
            ocp_bulk = pybamm.Scalar(0)
            dUdT = 0 * T
        elif self.reaction == "lead-acid main":
            c_e = variables[f"{Domain} electrolyte concentration [mol.m-3]"]
            # If c_e was broadcast, take only the orphan
            if isinstance(c_e, pybamm.Broadcast):
                c_e = c_e.orphans[0]
            ocp_surf = self.phase_param.U(c_e, self.param.T_init)
            dUdT = pybamm.Scalar(0)

            c_e_av = variables["X-averaged electrolyte concentration [mol.m-3]"]
            ocp_bulk = self.phase_param.U(c_e_av, self.param.T_init)

        elif self.reaction == "lead-acid oxygen":
            ocp_surf = self.param.U_Ox
            ocp_bulk = self.param.U_Ox
            dUdT = pybamm.Scalar(0)

        variables.update(self._get_standard_ocp_variables(ocp_surf, ocp_bulk, dUdT))
        return variables



import pybamm


def check_if_composite(options, electrode):
    if not isinstance(options, pybamm.BatteryModelOptions):
        options = pybamm.BatteryModelOptions(options)
    domain_options = getattr(options, electrode)
    particle_phases = domain_options["particle phases"]
    if particle_phases == "2":
        return True
    else:
        return False


def _has_hysteresis(electrode, options, phase=None):
    if not isinstance(options, pybamm.BatteryModelOptions):
        options = pybamm.BatteryModelOptions(options)
    hysteresis_options = [
        "current sigmoid",
        "one-state hysteresis",
        "one-state differential capacity hysteresis",
        # Also catch old names
        "Axen",
        "Wycisk",
    ]
    domain_options = getattr(options, electrode)
    if isinstance(domain_options["open-circuit potential"], str):
        return domain_options["open-circuit potential"] in hysteresis_options
    elif isinstance(domain_options["open-circuit potential"], tuple):
        ocp_option = domain_options["open-circuit potential"]
        if phase == "primary":
            return ocp_option[0] in hysteresis_options
        elif phase == "secondary":
            return ocp_option[1] in hysteresis_options
        else:
            return any(
                isinstance(item, str) and item in hysteresis_options
                for item in ocp_option
            )
    else:
        raise ValueError(
            f"Invalid open-circuit potential option: {domain_options['open-circuit potential']}"
        )


def _get_lithiation_delithiation(direction, electrode, options, phase=None):
    if direction is None or not _has_hysteresis(electrode, options, phase):
        return None
    elif (direction == "charge" and electrode == "negative") or (
        direction == "discharge" and electrode == "positive"
    ):
        return "lithiation"
    elif (direction == "discharge" and electrode == "negative") or (
        direction == "charge" and electrode == "positive"
    ):
        return "delithiation"
    else:
        raise ValueError()
